@@ -385,45 +385,22 @@ int hnbgw_cnlink_tx_ranap_reset_ack(struct hnbgw_cnlink *cnlink)
 	return hnbgw_cnlink_tx_sccp_unitdata_req(cnlink, msg);
 }
 
-static bool addr_has_pc_and_ssn(const struct osmo_sccp_addr *addr)
+/* Return address found in sccp address-book, and fill in missing fields in the
+ * entry with default values. */
+static struct osmo_ss7_instance *sccp_addr_by_name_filled(struct osmo_sccp_addr *dest, const char *addr_name, uint32_t default_pc)
 {
-	if (!(addr->presence & OSMO_SCCP_ADDR_T_SSN))
-		return false;
-	if (!(addr->presence & OSMO_SCCP_ADDR_T_PC))
-		return false;
-	return true;
-}
+	struct osmo_ss7_instance *s7i;
+	s7i = osmo_sccp_addr_by_name(dest, addr_name);
+	if (!s7i)
+		return NULL;
 
-static int resolve_addr_name(struct osmo_sccp_addr *dest, struct osmo_ss7_instance **ss7,
-			     const char *addr_name, const char *label,
-			     uint32_t default_pc)
-{
-	if (!addr_name) {
+	/* Address exists in address-book but may not be filled entirely: */
+	if (!dest->presence)
 		osmo_sccp_make_addr_pc_ssn(dest, default_pc, OSMO_SCCP_SSN_RANAP);
-		if (label)
-			LOGP(DCN, LOGL_INFO, "%s remote addr not configured, using default: %s\n", label,
-			     osmo_sccp_addr_name(*ss7, dest));
-		return 0;
-	}
+	else if (!(dest->presence & OSMO_SCCP_ADDR_T_SSN))
+		osmo_sccp_addr_set_ssn(dest, OSMO_SCCP_SSN_RANAP);
 
-	*ss7 = osmo_sccp_addr_by_name(dest, addr_name);
-	if (!*ss7) {
-		if (label)
-			LOGP(DCN, LOGL_ERROR, "%s remote addr: no such SCCP address book entry: '%s'\n",
-			     label, addr_name);
-		return -1;
-	}
-
-	osmo_sccp_addr_set_ssn(dest, OSMO_SCCP_SSN_RANAP);
-
-	if (!addr_has_pc_and_ssn(dest)) {
-		if (label)
-			LOGP(DCN, LOGL_ERROR, "Invalid/incomplete %s remote-addr: %s\n",
-			     label, osmo_sccp_addr_name(*ss7, dest));
-		return -1;
-	}
-
-	return 0;
+	return s7i;
 }
 
 char *hnbgw_cnlink_sccp_addr_to_str(struct hnbgw_cnlink *cnlink, const struct osmo_sccp_addr *addr)
@@ -451,13 +428,17 @@ static bool hnbgw_cnlink_sccp_cfg_changed(struct hnbgw_cnlink *cnlink)
 	bool changed = false;
 
 	if (cnlink->vty.remote_addr_name && cnlink->use.remote_addr_name) {
-		struct osmo_ss7_instance *ss7;
+		struct osmo_ss7_instance *s7i;
 		struct osmo_sccp_addr remote_addr = {};
 
 		/* Instead of comparing whether the address book entry names are different, actually resolve the
 		 * resulting SCCP address, and only restart the cnlink if the resulting address changed. */
-		resolve_addr_name(&remote_addr, &ss7, cnlink->vty.remote_addr_name, NULL, DEFAULT_PC_HNBGW);
-		if (osmo_sccp_addr_cmp(&remote_addr, &cnlink->remote_addr, OSMO_SCCP_ADDR_T_PC | OSMO_SCCP_ADDR_T_SSN))
+		s7i = sccp_addr_by_name_filled(&remote_addr, cnlink->vty.remote_addr_name,
+					       cnlink->pool->default_remote_pc);
+		if (!s7i)
+			return true;
+		if (osmo_sccp_addr_cmp(&remote_addr, &cnlink->remote_addr,
+				      OSMO_SCCP_ADDR_T_PC | OSMO_SCCP_ADDR_T_SSN))
 			changed = true;
 	} else if (cnlink->vty.remote_addr_name != cnlink->use.remote_addr_name) {
 		/* One of them is NULL, the other is not. */
@@ -465,7 +446,6 @@ static bool hnbgw_cnlink_sccp_cfg_changed(struct hnbgw_cnlink *cnlink)
 	}
 
 	/* if more cnlink configuration is added in the future, it needs to be compared here. */
-
 	return changed;
 }
 
@@ -485,19 +465,20 @@ static void hnbgw_cnlink_log_self(struct hnbgw_cnlink *cnlink)
  * Set cnlink->hnbgw_sccp_user to the new SCCP instance. Return 0 on success, negative on error. */
 int hnbgw_cnlink_start_or_restart(struct hnbgw_cnlink *cnlink)
 {
-	struct osmo_ss7_instance *ss7 = NULL;
+	struct osmo_ss7_instance *s7i = NULL;
 	struct hnbgw_sccp_user *hsu;
+	uint32_t ss7_id;
+	int rc;
 
 	/* If a hnbgw_sccp_user has already been set up, use that. */
 	if (cnlink->hnbgw_sccp_user) {
-		if (hnbgw_cnlink_sccp_cfg_changed(cnlink)) {
-			LOG_CNLINK(cnlink, DCN, LOGL_NOTICE, "config changed, restarting SCCP\n");
-			hnbgw_cnlink_drop_sccp(cnlink);
-		} else {
+		if (!hnbgw_cnlink_sccp_cfg_changed(cnlink)) {
 			LOG_CNLINK(cnlink, DCN, LOGL_DEBUG, "SCCP instance already set up, using %s\n",
 				   cnlink->hnbgw_sccp_user->name);
 			return 0;
 		}
+		LOG_CNLINK(cnlink, DCN, LOGL_NOTICE, "config changed, restarting SCCP\n");
+		hnbgw_cnlink_drop_sccp(cnlink);
 	} else {
 		LOG_CNLINK(cnlink, DCN, LOGL_DEBUG, "no SCCP instance selected yet\n");
 	}
@@ -505,49 +486,64 @@ int hnbgw_cnlink_start_or_restart(struct hnbgw_cnlink *cnlink)
 	/* Copy the current configuration: cnlink->use = cnlink->vty */
 	hnbgw_cnlink_cfg_copy(cnlink);
 
-	/* Figure out which cs7 instance to use. If cnlink->remote_addr_name is set, it points to an address book entry
-	 * in a specific cs7 instance. If it is not set, leave ss7 == NULL to use cs7 instance 0. */
-	if (cnlink->use.remote_addr_name) {
-		if (resolve_addr_name(&cnlink->remote_addr, &ss7, cnlink->use.remote_addr_name, cnlink->name,
-				      DEFAULT_PC_HNBGW)) {
+	if (!cnlink->use.remote_addr_name) {
+		/* No remote address configured in VTY, set a default one and
+		 * make sure it becomes registered in the sccp address-book: */
+		cnlink->use.remote_addr_name = talloc_strdup(cnlink, cnlink->pool->default_addr_name);
+		s7i = osmo_sccp_addr_by_name(&cnlink->remote_addr, cnlink->use.remote_addr_name);
+		if (!s7i) {
+			LOG_CNLINK(cnlink, DCN, LOGL_NOTICE, "To auto-configure cnlink, creating cs7 instance 0 implicitly\n");
+			s7i = osmo_ss7_instance_find_or_create(g_hnbgw, 0);
+			OSMO_ASSERT(s7i);
+			osmo_sccp_make_addr_pc_ssn(&cnlink->remote_addr,
+						   cnlink->pool->default_remote_pc,
+						   OSMO_SCCP_SSN_RANAP);
+			rc = osmo_sccp_addr_create(s7i, cnlink->use.remote_addr_name, &cnlink->remote_addr);
+			if (rc < 0) {
+				LOG_CNLINK(cnlink, DCN, LOGL_ERROR, "Failed adding address '%s' to sccp address-book!\n",
+					   cnlink->use.remote_addr_name);
+				return -EINVAL;
+			}
+		}
+		/* Update VTY config to show & point to the address dynamically added to address-book: */
+		cnlink->vty.remote_addr_name = talloc_strdup(cnlink, cnlink->use.remote_addr_name);
+	} else {
+		s7i = sccp_addr_by_name_filled(&cnlink->remote_addr, cnlink->use.remote_addr_name, cnlink->pool->default_remote_pc);
+		if (!s7i) {
 			LOG_CNLINK(cnlink, DCN, LOGL_ERROR, "cannot initialize SCCP: there is no SCCP address named '%s'\n",
 				   cnlink->use.remote_addr_name);
 			return -ENOENT;
 		}
-
 		LOG_CNLINK(cnlink, DCN, LOGL_DEBUG, "remote-addr is '%s', using cs7 instance %u\n",
-			   cnlink->use.remote_addr_name, osmo_ss7_instance_get_id(ss7));
-	} else {
-		/* If no address is configured, use the default remote CN address, according to legacy behavior. */
-		osmo_sccp_make_addr_pc_ssn(&cnlink->remote_addr, cnlink->pool->default_remote_pc, OSMO_SCCP_SSN_RANAP);
-	}
-
-	/* If no 'cs7 instance' has been selected by the address, see if there already is a cs7 0 we can use by default.
-	 * If it doesn't exist, it will get created by osmo_sccp_simple_client_on_ss7_id(). */
-	if (!ss7) {
-		ss7 = osmo_ss7_instance_find(0);
-		LOG_CNLINK(cnlink, DCN, LOGL_DEBUG, "Using default 'cs7 instance 0' (%s)\n", ss7 ? "already exists" : "will create");
-	}
-
-	if (ss7) {
-		/* Has another cnlink already set up an SCCP instance for this ss7? */
-		llist_for_each_entry(hsu, &g_hnbgw->sccp.users, entry) {
-			if (hsu->ss7 != ss7)
-				continue;
-			LOG_CNLINK(cnlink, DCN, LOGL_DEBUG, "using existing SCCP instance %s on cs7 instance %u\n",
-				hsu->name, osmo_ss7_instance_get_id(ss7));
-			cnlink->hnbgw_sccp_user = hsu;
-			hnbgw_sccp_user_get(cnlink->hnbgw_sccp_user, HSU_USE_CNLINK);
-			hnbgw_cnlink_log_self(cnlink);
-			return 0;
+			   cnlink->use.remote_addr_name, osmo_ss7_instance_get_id(s7i));
+		/* Address exists in address-book but may not be filled entirely: */
+		rc = osmo_sccp_addr_update(s7i, cnlink->use.remote_addr_name, &cnlink->remote_addr);
+		if (rc < 0) {
+			LOG_CNLINK(cnlink, DCN, LOGL_ERROR, "Failed updating address '%s' in sccp address-book!\n",
+				   cnlink->use.remote_addr_name);
+			return -EINVAL;
 		}
-		/* else cnlink->hnbgw_sccp_user stays NULL and is set up below. */
-		LOG_CNLINK(cnlink, DCN, LOGL_DEBUG, "cs7 instance %u has no configured SCCP instance yet\n", osmo_ss7_instance_get_id(ss7));
 	}
 
+	ss7_id = osmo_ss7_instance_get_id(s7i);
+
+	/* Has another cnlink already set up an SCCP instance for this s7i? */
+	llist_for_each_entry(hsu, &g_hnbgw->sccp.users, entry) {
+		if (hsu->ss7 != s7i)
+			continue;
+		LOG_CNLINK(cnlink, DCN, LOGL_DEBUG, "using existing SCCP instance %s on cs7 instance %u\n",
+			hsu->name, ss7_id);
+		cnlink->hnbgw_sccp_user = hsu;
+		hnbgw_sccp_user_get(cnlink->hnbgw_sccp_user, HSU_USE_CNLINK);
+		hnbgw_cnlink_log_self(cnlink);
+		return 0;
+	}
+	/* else cnlink->hnbgw_sccp_user stays NULL and is set up below. */
+	LOG_CNLINK(cnlink, DCN, LOGL_DEBUG, "cs7 instance %u has no configured SCCP instance yet\n", ss7_id);
+
 	/* No SCCP instance yet for this ss7. Create it. If no address name is given that resolves to a
 	 * particular cs7 instance above, use 'cs7 instance 0'. */
-	cnlink->hnbgw_sccp_user = hnbgw_sccp_user_alloc(ss7 ? osmo_ss7_instance_get_id(ss7) : 0);
+	cnlink->hnbgw_sccp_user = hnbgw_sccp_user_alloc(ss7_id);
 	hnbgw_sccp_user_get(cnlink->hnbgw_sccp_user, HSU_USE_CNLINK);
 	hnbgw_cnlink_log_self(cnlink);
 	return 0;
